import logging, re
from pprint import pformat
from django.conf import settings
from mockup import mockup
from dj_kunde.contrib.view_test.context_keeper import get_context

class TestGhostWriter:
    def __init__(self):
        self.url_filter = re.compile(settings.TEST_URL_FILTER)
        self.logger = logging.getLogger("test.write")

    def process_response(self, request, response):
        if self.url_filter.search(request.path):
            self.logger.debug("# --------------------------------------------------------------------")
            self.logger.debug("# %s %s" % (request.META["REQUEST_METHOD"], request.path))
            if request.GET or request.POST:
                self.logger.debug("GET = MultiValueDict(%s)", pformat(dict(request.GET.lists())))
                self.logger.debug("POST = MultiValueDict(%s)", pformat(dict(request.POST.lists())))
                self.logger.debug("response, context = runner.get_response(%r, %r, GET, POST)" %
                                  (request.path, request.META["REQUEST_METHOD"]))
            else:
                self.logger.debug("response, context = runner.get_response(%r, %r, {}, {})" %
                                  (request.path, request.META["REQUEST_METHOD"]))
            self.logger.debug("assert runner.check_response(response, %s)", pformat(mockup(response)))
            if response.status_code < 300:
                self.logger.debug("assert runner.check_context(response, context, %s)", pformat(get_context()))
        return response




"""This is a hack to pass the context around.
"""
# only exists in python 2.4+
try:
    from threading import local
except ImportError:
    # import copy of _thread_local.py from python 2.4
    from django.utils._threading_local import local
from mockup import mockup

_keeper = local()
_keeper.context = None

def set_context(context):
    _keeper.context = mockup(context)

def get_context():
    return _keeper.context



import types, os, tempfile
from pprint import pprint
from django.dispatch import dispatcher
from django.core import signals
from django.core.handlers.base import BaseHandler
from django.http import HttpRequest
from django.utils.datastructures import MergeDict
from dj_kunde.contrib.view_test.context_keeper import get_context, set_context

HEADER = "*" * 60

class TestRunner(object):
    def __init__(self):
        self.handler = BaseHandler()
        self.handler.load_middleware()
        self.cookies = {}

    def get_response(self, path, method, GET, POST):
        request = HttpRequest()
        request.path = path
        request.META['REQUEST_METHOD'] = method
        request.META['REMOTE_ADDR'] = '127.0.0.1'
        request.COOKIES = self.cookies
        request.GET = GET
        request.POST = POST
        request.REQUEST = MergeDict(GET, POST)
        set_context(None)

        dispatcher.send(signal=signals.request_started)
        response = self.handler.get_response(path, request)
        for middleware_method in self.handler._response_middleware:
            response = middleware_method(request, response)
        dispatcher.send(signal=signals.request_finished)

        context = get_context()
        self.cookies.update(dict([(k,v.value) for k,v in response.cookies.iteritems()]))
        return (response, context)

    def check_context(self, response, ist, soll):
        # remove all the keys that are not in soll:
        if ist == None:
            ist = {}
        #ist = dict([(k,v) for k,v in ist.items() if k in soll])
        if ist == soll:
            return True
        else:
            remove_ellipsis(ist, soll)
            if ist==soll:
                return True
            else:
                print "response is:"
                self.print_html(response)
                print HEADER
                print "context differs:"
                print HEADER
                print_diffs("context",ist,soll)
                #print "IST"
                #pprint (ist)
                #print "\nSOLL"
                #pprint (soll)
                return False

    def check_response(self, ist, soll):
        result = ist.status_code == soll["status_code"] and set(ist.cookies.keys()) == set(soll["cookies"].keys())
        if not result:
            print "response is:"
            self.print_html(ist)
            print HEADER
            if ist.status_code != soll["status_code"]:
                print "* status_code ist=%d, soll=%d" % (ist.status_code, soll["status_code"])
            if set(ist.cookies.keys()) != set(soll["cookies"].keys()):
                print "* cookies ist=%r, soll=%r" % (ist.cookies, soll["cookies"])
            print HEADER
        return result

    def print_html(self, response):
        fd, filename = tempfile.mkstemp('.html','viewtest-',text=True)
        f = os.fdopen(fd,"w")
        print "... writing html output to %s" % filename
        f.write(str(response))
        f.close()
        #print HEADER
        #f = popen("w3m -dump -T text/html","w")
        #if f:
            #f.write(str(response))
            #f.close()


def print_diffs(prefix, ist, soll):
    if ist==None and soll!=None:
        print "* %s: missing" % prefix
    else:
        for key, soll_value in soll.iteritems():
            ist_value = ist.get(key,None)
            if ist_value != soll_value:
                if isinstance(soll_value,dict) and isinstance(ist_value,dict):
                    print_diffs(prefix+"."+key, ist_value, soll_value)
                else:
                    print "* %s.%s:  ist=%r,\n* %s  soll=%r" % (prefix, key, ist_value, " " * (len(prefix)+len(key)+1), soll_value)
        for key, value in ist.iteritems():
            if key not in soll:
                print "* %s.%s: additional, ist=%r" % (prefix, key, value)


def remove_ellipsis(ist, soll):
    """if soll is a dict, remove all items with a value of `Ellipsis`, from both ist and soll.
    If soll is a list, remove all elemens that equal `Ellipsis`, from both ist and soll.
    Then apply this recursively to all elements of ist/soll

    >>> remove_ellipsis([1,2,3], [4,5, Ellipsis])
    ([1, 2], [4, 5])

    >>> remove_ellipsis([],[])
    ([], [])

    >>> remove_ellipsis([{"bla":1, "blo":"asdasd"}], [Ellipsis])
    ([], [])

    >>> remove_ellipsis({"eins":1, "zwei":2}, {"drei": 3, "eins": Ellipsis})
    ({'zwei': 2}, {'drei': 3})

    >>> remove_ellipsis([1,2,{3:3, 4:4, 5:[6,7]}], [0,1,{5:[Ellipsis,2]}])
    ([1, 2, {3: 3, 4: 4, 5: [7]}], [0, 1, {5: [2]}])

    >>> remove_ellipsis(1,Ellipsis)
    (1, Ellipsis)

    >>> remove_ellipsis(None, [Ellipsis])
    (None, [Ellipsis])
    """
    if isinstance(soll, dict) and isinstance(ist, dict):
        for key, value in soll.items():
            if value == Ellipsis:
                del soll[key]
                try:
                    del ist[key]
                except KeyError:
                    pass
            else:
                if isinstance(value, (dict,types.ListType, dict)) and  key in ist:
                    remove_ellipsis(ist[key], value)
    elif isinstance(soll, types.ListType) and isinstance(ist, types.ListType):
        deleteables = []
        for i, value in enumerate(soll):
            if value == Ellipsis:
                deleteables.insert(0,i)
            else:
                if isinstance(value, (dict,types.ListType, dict)):
                    remove_ellipsis(ist[i], value)
        for i in deleteables:
            del soll[i]
            del ist[i]
    return (ist,soll)

def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()




"""Hold the context in the session."""

from django import template
from django.conf import settings
from dj_kunde.contrib.view_test.context_keeper import set_context
register = template.Library()

class ContextKeeperNode(template.Node):
    def render(self, context):
        if settings.VIEW_TEST:
            set_context(context)
        return ''

@register.tag
def context_keeper(parser, tag):
    return ContextKeeperNode()




"""Create mockups of stuff.

    The idea is that eval(mockup(obj1)) == eval(mockup(obj2))
    is mostly equivalent to obj1==obj2.

    For containers, such as lists, dicts, sets, all components are
    'reprified'. For simple builtin types, it's the type itself, with the exception that all
    integers are normalized to longs. For other types, repr() is used.

    >>> mockup("abc")
    'abc'

    >>> mockup(range(5))
    [0, 1, 2, 3, 4]

    >>> import sys
    >>> d=mockup({"a": [type(1), type("s")], "b": set([Ellipsis])})
    >>> d["a"]
    ["<type 'int'>", "<type 'str'>"]
    >>> d["b"]
    set(['Ellipsis'])

    >>> from django.utils.datastructures import MultiValueDict
    >>> md = mockup(MultiValueDict({"a": [type(1), type("s")], "b": set([Ellipsis])}))
    >>> md
    MultiValueDict({'a': ["<type 'int'>", "<type 'str'>"], 'b': set(['Ellipsis'])})
    >>> md.getlist("a")
    ["<type 'int'>", "<type 'str'>"]
    >>> md["a"]
    "<type 'str'>"
"""

import types, re
from Cookie import BaseCookie
from django.utils.datastructures import MultiValueDict
from django.db.models.query import QuerySet
from django.forms import FormWrapper
from django.http import HttpResponse
from django.template.context import Context




# ignore these keys in context objects
context_ignore = ["LANGUAGES","LANGUAGE_CODE", "LANGUAGE_BIDI","sql_queries", "context", "debug", "perms", "MAILADMIN_ROOTURL", "AUTH_ROOTURL"]

# regulare expression, this is replaced in a repr() of unknown objects
mockup_others_re = re.compile(r" object at 0x[0-f]+>$")



def mockup(src):
    """returns a "mockup" for testing purposes. This is the main dispatcher.
    """
    return _dispatch.get(type(src), mockup_others)(src)

def mockup_others(src):
    """dispatches objects"""
    if isinstance(src, QuerySet):
        return mockup_list(src)
    elif isinstance(src, MultiValueDict):
        return mockup_multi_value_dict(src)
    elif isinstance(src, FormWrapper):
        return mockup_dict(src.__dict__)
    elif isinstance(src, BaseCookie):
        return mockup_cookie(src)
    elif isinstance(src, HttpResponse):
        return mockup_response(src)
    elif isinstance(src, Context):
        return mockup_context(src)
    else:
        return mockup_others_re.sub(" object>",repr(src))


def mockup_response(response):
    data = dict(status_code=response.status_code,
                cookies=response.cookies)
    if response.has_header("Location"):
        data["location"]=response["Location"]
    return mockup_dict(data)

def mockup_context(context):
    result = {}
    for d in context:
        result.update(d)
    for k in context_ignore:
        if k in result:
            del result[k]
        if "csrfmiddlewaretoken" in result:
            result["csrfmiddlewaretoken"] = "self.csrfmiddlewaretoken"
    return mockup_dict(result)


def identity(src):
    return src

def mockup_tuple(src):
    return tuple([mockup(x) for x in src])

def mockup_list(src):
    return [mockup(x) for x in src]

def mockup_dict(src):
    return dict([(k, mockup(v)) for k,v in src.iteritems()])

def mockup_multi_value_dict(src):
    return MultiValueDictMockup(dict([(k, mockup(v)) for k,v in src.lists()]))

def mockup_set(src):
    return set([mockup(x) for x in src])

def mockup_cookie(src):
    return CookieMockup(src)

class CookieMockup(object):
    def __init__(self, src):
        self.cookie = src
    def __repr__(self):
        return "%s(%r)" % (self.cookie.__class__.__name__, self.cookie.output([], header=""))

class MultiValueDictMockup(MultiValueDict):
    def __repr__(self):
        return "MultiValueDict(%s)" % dict.__repr__(self)

_dispatch = {
    types.IntType: identity,
    types.LongType: identity,
    types.StringType: identity,
    types.UnicodeType: identity,
    types.BooleanType: identity,
    types.FloatType: identity,
    types.NoneType: identity,
    types.TupleType: mockup_tuple,
    types.ListType: mockup_list,
    types.DictionaryType: mockup_dict,
    type(set()): mockup_set,
}


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()



